from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Protocol, TypeVar, cast

import awkward as ak
from awkward import Array as AwkwardArray
from awkward.forms import Form
from awkward.typetracer import typetracer_from_form, typetracer_with_report

from dask_awkward.layers.layers import (
    BackendT,
    ImplementsIOFunction,
    ImplementsNecessaryColumns,
)
from dask_awkward.lib.utils import (
    METADATA_ATTRIBUTES,
    FormStructure,
    buffer_keys_required_to_compute_shapes,
    form_with_unique_keys,
    parse_buffer_key,
    render_buffer_key,
    trace_form_structure,
    walk_graph_depth_first,
)

if TYPE_CHECKING:
    from awkward._nplikes.typetracer import TypeTracerReport

log = logging.getLogger(__name__)

T = TypeVar("T")


class ImplementsColumnProjectionMixin(ImplementsNecessaryColumns, Protocol):
    @property
    def form(self) -> Form:
        ...

    @property
    def behavior(self) -> dict | None:
        ...

    def project_columns(self: T, columns: frozenset[str]) -> T:
        ...

    def __call__(self, *args, **kwargs):
        ...


S = TypeVar("S", bound=ImplementsColumnProjectionMixin)


class ColumnProjectionMixin(ImplementsNecessaryColumns[FormStructure]):
    """A mixin to add column-centric buffer projection to an IO function.

    Classes that inherit from this mixin are assumed to be able to read at the
    granularity of _fields_ in a form. As such, the buffer projection is performed
    such that the smallest possible number of fields (columns) are read, even
    when only metadata buffers are required.
    """

    def mock(self: S) -> AwkwardArray:
        return cast(
            AwkwardArray, typetracer_from_form(self.form, behavior=self.behavior)
        )

    def mock_empty(self: S, backend: BackendT = "cpu") -> AwkwardArray:
        return cast(
            AwkwardArray,
            ak.to_backend(
                self.form.length_zero_array(highlevel=False, behavior=self.behavior),
                backend,
                highlevel=True,
            ),
        )

    def prepare_for_projection(
        self: S,
    ) -> tuple[AwkwardArray, TypeTracerReport, FormStructure]:
        form = form_with_unique_keys(self.form, "@")

        # Build typetracer and associated report object
        (meta, report) = typetracer_with_report(
            form,
            highlevel=True,
            behavior=self.behavior,
            buffer_key=render_buffer_key,
        )

        return (
            cast(AwkwardArray, meta),
            report,
            trace_form_structure(form, buffer_key=render_buffer_key),
        )

    def necessary_columns(
        self: S,
        report: TypeTracerReport,
        state: FormStructure,
    ) -> frozenset[str]:
        ## Read from stash
        # Form hierarchy information
        form_key_to_parent_form_key = state["form_key_to_parent_form_key"]
        form_key_to_child_form_keys: dict[str, list[str]] = {}
        for child_key, parent_key in form_key_to_parent_form_key.items():
            form_key_to_child_form_keys.setdefault(parent_key, []).append(child_key)  # type: ignore
        form_key_to_form = state["form_key_to_form"]
        # Buffer hierarchy information
        form_key_to_buffer_keys = state["form_key_to_buffer_keys"]
        # Column hierarchy information
        form_key_to_path = state["form_key_to_path"]

        # Require the data of metadata buffers above shape-only requests
        data_buffers = {
            *report.data_touched,
            *buffer_keys_required_to_compute_shapes(
                parse_buffer_key,
                report.shape_touched,
                form_key_to_parent_form_key,
                form_key_to_buffer_keys,
            ),
        }

        # We can't read buffers directly, but if we encounter a metadata
        # buffer, then we should be able to pick any child.
        paths = set()
        wildcard_form_key = set()
        for buffer_key in data_buffers:
            form_key, attribute = parse_buffer_key(buffer_key)
            if attribute in METADATA_ATTRIBUTES:
                wildcard_form_key.add(form_key)
            else:
                paths.add(form_key_to_path[form_key])

        # Select the most appropriate column for each wildcard
        for form_key in wildcard_form_key:
            # Find (DFS) any non-empty record form in any child
            recursive_child_forms = (
                form_key_to_form[k]
                for k in walk_graph_depth_first(form_key, form_key_to_child_form_keys)
            )
            record_form_keys_with_contents = (
                f.form_key
                for f in recursive_child_forms
                if isinstance(f, ak.forms.RecordForm) and f.contents
            )
            # Now find the deepest of such records
            try:
                last_record_form_key = next(record_form_keys_with_contents)
            except StopIteration:
                # This is a leaf! Therefore, we read this column
                paths.add(form_key_to_path[form_key])
                continue
            else:
                # Ensure we get the "actual" last form key
                for last_record_form_key in record_form_keys_with_contents:
                    ...

            # First see if any child is already included
            for any_child_form_key in form_key_to_child_form_keys[last_record_form_key]:
                any_child_path = form_key_to_path[any_child_form_key]
                if any_child_path in paths:
                    break
            # Otherwise, add the last child
            else:
                paths.add(any_child_path)
        return frozenset({".".join(p) for p in paths if p})

    def project(
        self: S,
        report: TypeTracerReport,
        state: FormStructure,
    ) -> ImplementsIOFunction:
        if not self.use_optimization:  # type: ignore[attr-defined]
            return self

        return self.project_columns(self.necessary_columns(report, state))



from __future__ import annotations

import copy
import logging
import warnings
from collections.abc import Hashable, Iterable, Mapping
from typing import TYPE_CHECKING, Any, cast

import dask.config
from awkward.typetracer import touch_data
from dask.blockwise import fuse_roots, optimize_blockwise
from dask.core import flatten
from dask.highlevelgraph import HighLevelGraph
from dask.local import get_sync

from dask_awkward.layers import AwkwardBlockwiseLayer, AwkwardInputLayer

if TYPE_CHECKING:
    from awkward._nplikes.typetracer import TypeTracerReport

log = logging.getLogger(__name__)

COLUMN_OPT_FAILED_WARNING_MSG = """The necessary columns optimization failed; exception raised:

{exception} with message {message}.

Please see the FAQ section of the docs for more information:
https://dask-awkward.readthedocs.io/en/stable/more/faq.html

"""


def all_optimizations(
    dsk: Mapping,
    keys: Hashable | list[Hashable] | set[Hashable],
    **_: Any,
) -> Mapping:
    """Run all optimizations that benefit dask-awkward computations.

    This function will run both dask-awkward specific and upstream
    general optimizations from core dask.

    """
    if not isinstance(keys, (list, set)):
        keys = (keys,)  # pragma: no cover
    keys = tuple(flatten(keys))

    if not isinstance(dsk, HighLevelGraph):
        dsk = HighLevelGraph.from_collections(str(id(dsk)), dsk, dependencies=())

    else:
        # Perform dask-awkward specific optimizations.
        dsk = optimize(dsk, keys=keys)

        # Perform Blockwise optimizations for HLG input
        dsk = optimize_blockwise(dsk, keys=keys)
        # fuse nearby layers
        dsk = fuse_roots(dsk, keys=keys)  # type: ignore

    # cull unncessary tasks
    dsk = dsk.cull(set(keys))  # type: ignore

    return dsk


def optimize(
    dsk: HighLevelGraph,
    keys: Hashable | list[Hashable] | set[Hashable],
    **_: Any,
) -> Mapping:
    """Run optimizations specific to dask-awkward.

    This is currently limited to determining the necessary columns for
    input layers.

    """
    if dask.config.get("awkward.optimization.enabled"):
        which = dask.config.get("awkward.optimization.which")
        if "columns" in which:
            dsk = optimize_columns(dsk)
        if "layer-chains" in which:
            dsk = rewrite_layer_chains(dsk, keys)

    return dsk


def _prepare_buffer_projection(
    dsk: HighLevelGraph,
) -> tuple[dict[str, TypeTracerReport], dict[str, Any]] | None:
    """Pair layer names with lists of necessary columns."""
    import awkward as ak

    if not _has_projectable_awkward_io_layer(dsk):
        return None

    layer_to_projection_state: dict[str, Any] = {}
    layer_to_reports: dict[str, TypeTracerReport] = {}
    projection_layers = dict(dsk.layers)

    for name, lay in dsk.layers.items():
        if isinstance(lay, AwkwardInputLayer):
            if lay.is_projectable:
                # Insert mocked array into layers, replacing generation func
                # Keep track of mocked state
                (
                    projection_layers[name],
                    layer_to_reports[name],
                    layer_to_projection_state[name],
                ) = lay.prepare_for_projection()
            elif lay.is_mockable:
                projection_layers[name] = lay.mock()
        elif hasattr(lay, "mock"):
            projection_layers[name] = lay.mock()

    for name in _ak_output_layer_names(dsk):
        projection_layers[name] = _mock_output(projection_layers[name])

    hlg = HighLevelGraph(projection_layers, dsk.dependencies)

    # this loop builds up what are the possible final leaf nodes by
    # inspecting the dependents dictionary. If something does not have
    # a dependent, it must be the end of a graph. These are the things
    # we need to compute for; we only use a single partition (the
    # first). for a single collection `.compute()` this list will just
    # be length 1; but if we are using `dask.compute` to pass in
    # multiple collections to be computed simultaneously, this list
    # will increase in length.
    leaf_layers_keys = [
        (k, 0) for k, v in dsk.dependents.items() if isinstance(v, set) and len(v) == 0
    ]

    # now we try to compute for each possible output layer key (leaf
    # node on partition 0); this will cause the typetacer reports to
    # get correct fields/columns touched. If the result is a record or
    # an array we of course want to touch all of the data/fields.
    try:
        for layer in hlg.layers.values():
            layer.__dict__.pop("_cached_dict", None)
        results = get_sync(hlg, leaf_layers_keys)
        for out in results:
            if isinstance(out, (ak.Array, ak.Record)):
                touch_data(out)
    except Exception as err:
        on_fail = dask.config.get("awkward.optimization.on-fail")
        # this is the default, throw a warning but skip the optimization.
        if on_fail == "warn":
            warnings.warn(
                COLUMN_OPT_FAILED_WARNING_MSG.format(exception=type(err), message=err)
            )
        # option "pass" means do not throw warning but skip the optimization.
        elif on_fail == "pass":
            log.debug("Column projection optimization failed; optimization skipped.")
        # option "raise" to raise the exception here
        elif on_fail == "raise":
            raise
        else:
            raise ValueError(
                f"Invalid awkward.optimization.on-fail option: {on_fail}.\n"
                "Valid options are 'warn', 'pass', or 'raise'."
            )
        return None
    else:
        return layer_to_reports, layer_to_projection_state


def optimize_columns(dsk: HighLevelGraph) -> HighLevelGraph:
    """Run column projection optimization.

    This optimization determines which columns from an
    ``AwkwardInputLayer`` are necessary for a complete computation.

    For example, if a parquet dataset is loaded with fields:
    ``["foo", "bar", "baz.x", "baz.y"]``

    And the following task graph is made:

    >>> ds = dak.from_parquet("/path/to/dataset")
    >>> z = ds["foo"] - ds["baz"]["y"]

    Upon calling z.compute() the AwkwardInputLayer created in the
    from_parquet call will only read the parquet columns ``foo`` and
    ``baz.y``.

    Parameters
    ----------
    dsk : HighLevelGraph
        Task graph to optimize.

    Returns
    -------
    HighLevelGraph
        New, optimized task graph with column-projected ``AwkwardInputLayer``.

    """
    projection_data = _prepare_buffer_projection(dsk)
    if projection_data is None:
        return dsk

    # Unpack result
    layer_to_reports, layer_to_projection_state = projection_data

    # Project layers using projection state
    layers = dict(dsk.layers)
    for name, state in layer_to_projection_state.items():
        layers[name] = cast(AwkwardInputLayer, layers[name]).project(
            report=layer_to_reports[name], state=state
        )

    return HighLevelGraph(layers, dsk.dependencies)


def _layers_with_annotation(dsk: HighLevelGraph, key: str) -> list[str]:
    return [n for n, v in dsk.layers.items() if (v.annotations or {}).get(key)]


def _ak_output_layer_names(dsk: HighLevelGraph) -> list[str]:
    """Get a list output layer names.

    Output layer names are annotated with 'ak_output'.

    Parameters
    ----------
    dsk : HighLevelGraph
        Graph of interest.

    Returns
    -------
    list[str]
        Names of the output layers.

    """
    return _layers_with_annotation(dsk, "ak_output")


def _has_projectable_awkward_io_layer(dsk: HighLevelGraph) -> bool:
    """Check if a graph at least one AwkwardInputLayer that is project-able."""
    for _, v in dsk.layers.items():
        if isinstance(v, AwkwardInputLayer) and v.is_projectable:
            return True
    return False


def _touch_all_data(*args, **kwargs):
    """Mock writing an ak.Array to disk by touching data buffers."""
    for arg in args + tuple(kwargs.values()):
        touch_data(arg)


def _mock_output(layer):
    """Update a layer to run the _touch_all_data."""
    assert len(layer.dsk) == 1

    new_layer = copy.deepcopy(layer)
    mp = new_layer.dsk.copy()
    for k in iter(mp.keys()):
        mp[k] = (_touch_all_data,) + mp[k][1:]
    new_layer.dsk = mp
    return new_layer


def rewrite_layer_chains(dsk: HighLevelGraph, keys: Any) -> HighLevelGraph:
    """Smush chains of blockwise layers into a single layer.

    The logic here identifies chains by popping layers (in arbitrary
    order) from a set of all layers in the task graph and walking
    through the dependencies (parent layers) and dependents (child
    layers). If a multi layer chain is discovered we compress it into
    a single layer with the second loop below (for chain in chains;
    that step rewrites the graph). In the chain building logic, if a
    layer exists in the `keys` argument (the keys necessary for the
    compute that we are optimizing for), we shortcircuit the logic to
    ensure we do not chain layers that contain a necessary key inside
    (these layers are called `required_layers` below).

    Parameters
    ----------
    dsk : HighLevelGraph
        Task graph to optimize.
    keys : Any
        Keys that are requested by the compute that is being
        optimized.

    Returns
    -------
    HighLevelGraph
        New, optimized task graph.

    """
    # dask.optimization.fuse_liner for blockwise layers
    import copy

    chains = []
    deps = copy.copy(dsk.dependencies)

    required_layers = {k[0] for k in keys}
    layers = {}
    # find chains; each chain list is at least two keys long
    dependents = dsk.dependents
    all_layers = set(dsk.layers)
    while all_layers:
        lay = all_layers.pop()
        val = dsk.layers[lay]
        if not isinstance(val, AwkwardBlockwiseLayer):
            # shortcut to avoid making comparisons
            layers[lay] = val  # passthrough unchanged
            continue
        children = dependents[lay]
        chain = [lay]
        lay0 = lay
        while (
            len(children) == 1
            and dsk.dependencies[list(children)[0]] == {lay}
            and isinstance(dsk.layers[list(children)[0]], AwkwardBlockwiseLayer)
            and len(dsk.layers[lay]) == len(dsk.layers[list(children)[0]])
            and lay not in required_layers
        ):
            # walk forwards
            lay = list(children)[0]
            chain.append(lay)
            all_layers.remove(lay)
            children = dependents[lay]
        lay = lay0
        parents = dsk.dependencies[lay]
        while (
            len(parents) == 1
            and dependents[list(parents)[0]] == {lay}
            and isinstance(dsk.layers[list(parents)[0]], AwkwardBlockwiseLayer)
            and len(dsk.layers[lay]) == len(dsk.layers[list(parents)[0]])
            and list(parents)[0] not in required_layers
        ):
            # walk backwards
            lay = list(parents)[0]
            chain.insert(0, lay)
            all_layers.remove(lay)
            parents = dsk.dependencies[lay]
        if len(chain) > 1:
            chains.append(chain)
            layers[chain[-1]] = copy.copy(
                dsk.layers[chain[-1]]
            )  # shallow copy to be mutated
        else:
            layers[lay] = val  # passthrough unchanged

    # do rewrite
    for chain in chains:
        # inputs are the inputs of chain[0]
        # outputs are the outputs of chain[-1]
        # .dsk is composed from the .dsk of each layer
        outkey = chain[-1]
        layer0 = dsk.layers[chain[0]]
        outlayer = layers[outkey]
        numblocks = [nb[0] for nb in layer0.numblocks.values() if nb[0] is not None][0]  # type: ignore
        deps[outkey] = deps[chain[0]]  # type: ignore
        [deps.pop(ch) for ch in chain[:-1]]  # type: ignore

        subgraph = layer0.dsk.copy()  # type: ignore
        indices = list(layer0.indices)  # type: ignore
        parent = chain[0]

        outlayer.io_deps = layer0.io_deps  # type: ignore
        for chain_member in chain[1:]:
            layer = dsk.layers[chain_member]
            for k in layer.io_deps:  # type: ignore
                outlayer.io_deps[k] = layer.io_deps[k]  # type: ignore
            func, *args = layer.dsk[chain_member]  # type: ignore
            args2 = _recursive_replace(args, layer, parent, indices)
            subgraph[chain_member] = (func,) + tuple(args2)
            parent = chain_member
        outlayer.numblocks = {i[0]: (numblocks,) for i in indices if i[1] is not None}  # type: ignore
        outlayer.dsk = subgraph  # type: ignore
        if hasattr(outlayer, "_dims"):
            del outlayer._dims
        outlayer.indices = tuple(  # type: ignore
            (i[0], (".0",) if i[1] is not None else None) for i in indices
        )
        outlayer.output_indices = (".0",)  # type: ignore
        outlayer.inputs = getattr(layer0, "inputs", set())  # type: ignore
        if hasattr(outlayer, "_cached_dict"):
            del outlayer._cached_dict  # reset, since original can be mutated
    return HighLevelGraph(layers, deps)


def _recursive_replace(args, layer, parent, indices):
    args2 = []
    for arg in args:
        if isinstance(arg, str) and arg.startswith("__dask_blockwise__"):
            ind = int(arg[18:])
            if layer.indices[ind][1] is None:
                # this is a simple arg
                args2.append(layer.indices[ind][0])
            elif layer.indices[ind][0] == parent:
                # arg refers to output of previous layer
                args2.append(parent)
            else:
                # arg refers to things defined in io_deps
                indices.append(layer.indices[ind])
                args2.append(f"__dask_blockwise__{len(indices) - 1}")
        elif isinstance(arg, list):
            args2.append(_recursive_replace(arg, layer, parent, indices))
        elif isinstance(arg, tuple):
            args2.append(tuple(_recursive_replace(arg, layer, parent, indices)))
        # elif isinstance(arg, dict):
        else:
            args2.append(arg)
    return args2


def _buffer_keys_for_layer(
    buffer_keys: Iterable[str], known_buffer_keys: frozenset[str]
) -> set[str]:
    return {k for k in buffer_keys if k in known_buffer_keys}
